#include "window.h"

#include <iostream>
#include <cstdlib>
#include <cstring>

#include "glerror.h"

static const char* kSpriteVertexShader="Sprite.vertex.glsl";
static const char* kSpriteShader="Sprite.fragment.glsl";

static const char* kStarTextureFile="star1.tga";

const int kStars = 1536;

Window::Window(const char * title, int width, int height){
    title_ = title;
    width_ = width;
    height_ = height;
    fullscreen_ = false;
}

void Window::Initialize(int major_gl_version, int minor_gl_version){

    InitGlfwOrDie(major_gl_version, minor_gl_version);
    InitGlewOrDie();


    std::cout << "OpenGL initialized: OpenGL version: " << glGetString(GL_VERSION) << " GLSL version: " << glGetString(GL_SHADING_LANGUAGE_VERSION) << std::endl;

    InitTextures();
    InitModels();
    InitPrograms();

    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
    glClearColor(0.0f, 0.0f, 0.0f, 1.0f);

}

void Window::InitGlfwOrDie(int major_gl_version, int minor_gl_version){
    if ( !glfwInit() ) {
        std::cerr << "ERROR: Could not initialize GLFW" << std::endl;
        exit(EXIT_FAILURE);
    }


    glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, major_gl_version);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, minor_gl_version);
#ifdef DEBUG
    glfwWindowHint(GLFW_OPENGL_DEBUG_CONTEXT, GLFW_TRUE);
#endif
    window_ = glfwCreateWindow(width_, height_, title_, nullptr, nullptr);
    if (!window_){
        std::cerr << "ERROR: Could not create a new rendering window" << std::endl;
        glfwTerminate();
        exit(EXIT_FAILURE);
    }
    glfwMakeContextCurrent(window_);
}

void Window::InitGlewOrDie(){
    GLenum glew_init_result;
    glewExperimental = GL_TRUE;
    glew_init_result = glewInit();

    if (GLEW_OK != glew_init_result) {
        std::cerr << "Glew ERROR: " << glewGetErrorString(glew_init_result) << std::endl;
        glfwTerminate();
        exit(EXIT_FAILURE);
    }
#ifdef DEBUG
    if(glDebugMessageCallback){
        std::cout << "Register OpenGL debug callback " << std::endl;
        glEnable(GL_DEBUG_OUTPUT_SYNCHRONOUS);
        glDebugMessageCallback((GLDEBUGPROC) OpenglCallbackFunction, NULL);
        GLuint unused_ids = 0;
        glDebugMessageControl(GL_DONT_CARE,
            GL_DONT_CARE,
            GL_DONT_CARE,
            0,
            &unused_ids,
            GL_FALSE);
    }
    else
        std::cout << "glDebugMessageCallback not available" << std::endl;
#endif

}


void Window::InitModels(){
    stars_.Initialize(kStars);
    stars_.SetTexture(star_texture_);
    stars_.SetTextureUnit(GL_TEXTURE0);

}

void Window::InitTextures(){
    star_texture_.Initialize(kStarTextureFile);
}

void Window::InitPrograms(){
    projection_matrix_ = Mat4::CreatePerspectiveProjectionMatrix(60,
                static_cast<GLfloat>(width_)/static_cast<GLfloat>(height_), 0.1f, 1000.0f);

    //
    program_.Initialize(kSpriteVertexShader, kSpriteShader);
    glUseProgram(program_);
    program_.SetTextureUnit(0);
    program_.SetProjectionMatrix(projection_matrix_);

    glUseProgram(0);
}

void Window::SetProjectionMatrix(){
    glUseProgram(program_);
    program_.SetProjectionMatrix(projection_matrix_);
    glUseProgram(0);
}


//////////////////////////////////////////////////////////////

void Window::Resize(int new_width, int new_height){
    width_ = new_width;
    height_ = new_height;
    projection_matrix_ = Mat4::CreatePerspectiveProjectionMatrix(60, (float)width_/(float)height_, 0.1f, 1000.0f);
    SetProjectionMatrix();
    glViewport(0, 0, width_, height_);
}

void Window::KeyEvent(int key, int /*scancode*/, int action, int /*mods*/){
    if(action == GLFW_PRESS){
        switch (key){
            case GLFW_KEY_ESCAPE:
                glfwSetWindowShouldClose(window_, GLFW_TRUE);
            break;
            case GLFW_KEY_F11:
               fullscreen_ = not fullscreen_;
               if(fullscreen_){
                  glfwSetWindowMonitor(window_, glfwGetPrimaryMonitor(), 0, 0, width_, height_, GLFW_DONT_CARE);
               }
               else{
                  glfwSetWindowMonitor(window_, nullptr, 0, 0, width_, height_, GLFW_DONT_CARE);
               }
            break;
        }
// glfwSetWindowMonitor(window, _fullscreen ? glfwGetPrimaryMonitor() : NULL, 0, 0, _width, _height, GLFW_DONT_CARE);        
    }
}

void Window::Run(void){
    while (!glfwWindowShouldClose(window_)){
         glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

         stars_.Move(0.1f*static_cast<float>(clock())/CLOCKS_PER_SEC);
         stars_.Draw(program_);
        
         glfwSwapBuffers(window_);
         glfwPollEvents();
    }

}

